import logging
from pathlib import Path
from string import Template

from programlib import language_

from seidr.gpt import explore_gpt


def initial_prompt(task_description, examples):
    prompt = task_description
    prompt += '\nFor example,'
    for sample_inputs, sample_outputs in examples:
        prompt += ''
        prompt += '\ninput:\n'
        for sample_input in sample_inputs:
            prompt += sample_input + '\n'
        prompt += 'output:'
        for sample_output in sample_outputs:
            prompt += '\n' + sample_output
    return prompt

def gpt_assisted_prompt(debug_prompt_text, task_description, input, expected_output, actual_output):
    """
    Create description of a bug using GPT3 completion.
    """
    assert 'GPT ---' in debug_prompt_text and len(debug_prompt_text.split(' --- ')) >= 3, 'Invalid prompt'
    _, code_behavior, debug_prompt_text = debug_prompt_text.split(' --- ')
    # Form problem description using template
    code_behavior = code_behavior.format(
        t=task_description,
        i=input,
        o=expected_output,
        a=actual_output)

    # Get GPT summary of a bug
    logging.info(f'\nCode behavior:\n{code_behavior}')
    bug_description = next(explore_gpt(source=code_behavior, instruction=None,
                                       modality='text', batch_size=1, t=0.0, delta_t=0.2))
    # Form debug prompt using template
    debug_prompt_text = debug_prompt_text.format(
        s=bug_description,
        t=task_description,
        i=input,
        o=expected_output,
        a=actual_output)
    return debug_prompt_text


def write_debug_prompt(test_runs, debug_prompt_text, task_description=None):
    logging.info('Updating debug prompt')
    mistake = min(test_runs, key=lambda run: run.correctness)
    if len(mistake) > 0:
        if mistake.error_lines:
            error_lines = '\n'.join(mistake.error_lines)
            return f'Fix {error_lines}'
        else:
            i = '\\n'.join(mistake.input_lines)
            o = '\\n'.join(mistake.expected_output_lines)
            if 'GPT ---' in debug_prompt_text:
                output_lines = '\n'.join([s.decode("utf-8") if type(s) == bytes else s for s in mistake.output_lines])
                return gpt_assisted_prompt(
                    debug_prompt_text, task_description, mistake.input_lines,
                    mistake.expected_output_lines, output_lines)
            return debug_prompt_text.format(i=i, o=o)
    else:
        logging.info('\n\nrun.correctness = 1 for all runs, mistake lines are empty\n\n')
        return f'Do not change anything'


def start_coding(prompt, language='C++'):
    language = language_(language)
    template_name = language.source.format(name='template')

    with open(Path(__file__).parent / 'code-templates' / template_name) as f:
        template = Template(f.read())

    return template.substitute(prompt=prompt)
